program test_laplacian_helmholtz

  !! Semi-implicit semi-Lagrangian shallow water model using double Fourier series

  use prm_phconst, only : PI, PI2, ER, OMG, GRAV
  use fft, only : fft__ini, fft__towave, fft__togrid
  use fft_y, only : fft_y__ini, fft_y__g2w, fft_y__g2w_uv, fft_y__w2g, fft_y__w2g_dy, fft_y__w2g_uv
  use xderiv, only : xderiv_run
  use com_dfs, only : com_dfs__ini
  use uv2divrot_dfs, only : uv2divrot_dfs__run, uv2divrot_dfs__uv2chipsi, uv2divrot_dfs__laplacian
  use divrot2uv_dfs, only : divrot2uv_dfs__run
  use helmholtz_dfs, only : helmholtz_dfs__run, helmholtz_dfs__like, helmholtz_dfs__hdiff2
  use hdiff4_dfs, only : hdiff4_dfs__run
  use monit, only : monit__ini, monit__output
  use monit_spectrum, only : monit_spectrum__ini, monit_spectrum__output
  use monit_norm, only : monit_norm__ini, monit_norm__output
  use grads, only : grads__outxy
  use e_time, only : e_time__start, e_time__end, e_time__output
!
  implicit none
!
!  integer,parameter :: JCN_DFS = 0 !! Old method and Cheong's basis functions
!  integer,parameter :: JCN_DFS = 1 !! New method to calculate expantion coefficients
  integer,parameter :: JCN_DFS = 2 !! New method to calculate expantion coefficients and New basis functions

  integer,parameter :: JCN_GRID = 0   !! colat=PI*(j+0.5)/JMAX0, j=0,...,JMAX0-1
!  integer,parameter :: JCN_GRID = 1   !! colat=PI*j/JMAX0, j=0,...,JMAX0   (JMAX=JMAX0+1)
!  integer,parameter :: JCN_GRID = -1  !! colat=PI*j/JMAX0, j=1,...,JMAX0-1 (JMAX=JMAX0-1)
  
!  integer,parameter :: JCN_ZONALFILTER = 0 !! No zonal filter
  integer,parameter :: JCN_ZONALFILTER = 1 !! Use zonal Fourier filter
  integer,parameter :: M0 = 20             !! For zonal Fourier filter

!  integer,parameter :: IMAX = 16
!  integer,parameter :: JMAX = 8 + JCN_GRID
!  integer,parameter :: NMAX = 7
!  real(8),save :: TIMESTEP = 3600.0d0         !! 1 hour
!
  integer,parameter :: IMAX = 128            !! About 300km resolution
  integer,parameter :: JMAX = 64+JCN_GRID
!  integer,parameter :: NMAX = 63             !! Linear grid
  integer,parameter :: NMAX = 42            !! Quadric grid
!
!  integer,parameter :: IMAX = 160            !! 250km resolution
!  integer,parameter :: JMAX = 80+JCN_GRID
!  integer,parameter :: NMAX = 79             !! Linear grid
!  integer,parameter :: NMAX = 53             !! Quadric grid
!
!  integer,parameter :: IMAX = 320            !! 120km resolution
!  integer,parameter :: JMAX = 160+JCN_GRID
!  integer,parameter :: NMAX = 159            !! Linear grid
!  integer,parameter :: NMAX = 106            !! Quadric grid
!
!  integer,parameter :: IMAX = 640            !! 60km resolution
!  integer,parameter :: JMAX = 320+JCN_GRID
!  integer,parameter :: NMAX = 319            !! Linear grid
!  integer,parameter :: NMAX = 213            !! Quadric grid
!
!  integer,parameter :: IMAX = 1920           !! 20km resolution
!  integer,parameter :: JMAX = 960+JCN_GRID   !! J is from North to South
!  integer,parameter :: NMAX = 959            !! Linear grid
!  integer,parameter :: NMAX = 639            !! Quadric grid
!
!  integer,parameter :: IMAX = 3840           !! 10km resolution
!  integer,parameter :: JMAX = 1920+JCN_GRID  !! J is from North to South
!  integer,parameter :: NMAX = 1919           !! Linear grid
!  integer,parameter :: NMAX = 1279           !! Quadric grid
!  integer,parameter :: NMAX = 959            !! Cubic grid
!
!  integer,parameter :: IMAX = 7680           !! 5km resolution
!  integer,parameter :: JMAX = 3840+JCN_GRID  !! J is from North to South
!  integer,parameter :: NMAX = 3839           !! Linear grid
!  integer,parameter :: NMAX = 2559           !! Quadric grid
!  integer,parameter :: NMAX = 1919           !! Cubic grid
!
!  integer,parameter :: IMAX = 15360          !! 2.6km resolution
!  integer,parameter :: JMAX = 7680+JCN_GRID  !! J is from North to South
!  integer,parameter :: NMAX = 7679           !! Linear grid
!  integer,parameter :: NMAX = 5119           !! Quadric grid
!  integer,parameter :: NMAX = 3839           !! Cubic grid
!
!  integer,parameter :: IMAX = 20480          !! 2.0km resolution
!  integer,parameter :: JMAX = 10240+JCN_GRID !! J is from North to South
!  integer,parameter :: NMAX = 10239          !! Linear grid
!  integer,parameter :: NMAX = 6826           !! Quadric grid
!  integer,parameter :: NMAX = 5119           !! Cubic grid
!
!  integer,parameter :: IMAX = 30720          !! 1.3km resolution
!  integer,parameter :: JMAX = 15360+JCN_GRID !! J is from North to South
!  integer,parameter :: NMAX = 15359          !! Linear grid   
!  integer,parameter :: NMAX = 10239          !! Quadric grid
!  integer,parameter :: NMAX = 7679           !! Cubic grid

  integer,parameter :: MMAX = NMAX
  integer,parameter :: NNUM = NMAX + 1
!
  integer,parameter :: IMAX2=IMAX/2 !! IMAX must be even.
!
  real(8),parameter :: DLON = PI2/IMAX
  real(8),parameter :: DLAT = PI/(JMAX-JCN_GRID)
!
  real(8),save :: ALON(IMAX)
  real(8),save :: ALON1(IMAX)
  real(8),save :: SINLON(IMAX)
  real(8),save :: COSLON(IMAX)
  real(8),save :: ALAT(JMAX)
  real(8),save :: YLAT(JMAX)
  real(8),save :: COSLAT(JMAX)
  real(8),save :: COSLAT_INV(JMAX)
  real(8),save :: SINLAT(JMAX)
!
  real(8),save :: ACOS2LAT_INV(JMAX)
  real(8),save :: ACOSLAT_INV(JMAX)
  real(8),save :: WEIGHT(JMAX)
  
  real(8),save :: WGAUSS(7)
  real(8),save :: XGAUSS(7)
!
! ==================================================================
!
  call e_time__start(1,"test_laplacian")

  call initialize
!
  call main
!
  call e_time__end(1,"test_laplacian")
!  call e_time__output
!
!
!******************************************************************
contains
!******************************************************************


subroutine initialize
!
  integer :: i,j
!
! -----------------------------------------------------------------------------
!  
  write(6,*) "JCN_DFS =",JCN_DFS
  write(6,*) "JCN_GRID=",JCN_GRID
  write(6,*) "IMAX    =",IMAX
  write(6,*) "JMAX    =",JMAX
  write(6,*) "NMAX    =",NMAX
  write(6,*) "JCN_ZONALFILTER=",JCN_ZONALFILTER
  
  call com_dfs__ini( JCN_DFS, JCN_GRID, IMAX, JMAX, MMAX, NMAX )
  call fft_y__ini( IMAX, JMAX, MMAX, alat, weight ) !IN
  
!$OMP PARALLEL default(SHARED), private(j,i)
 !$OMP DO schedule(STATIC)
  do j=1,JMAX
    COSLAT(j) = cos(ALAT(j))
    SINLAT(j) = sin(ALAT(j))
    COSLAT_INV(j) = 1.0d0/COSLAT(j)
    YLAT(j) = ALAT(j)*180.0d0/PI
    ACOS2LAT_INV(j) = COSLAT_INV(j)**2/ER
    ACOSLAT_INV(j) = COSLAT_INV(j)/ER
  end do
 !$OMP END DO

 !$OMP DO schedule(STATIC)
  do i=1,IMAX
    ALON(i) = DLON*(i-1)     !##   0 <= ALON  < 2*PI
    if ( ALON(i) >= PI ) then
       ALON1(i) = ALON(i)-2*PI  !## -PI <= ALON1 < PI
    else
       ALON1(i) = ALON(i)       !## -PI <= ALON1 < PI
    end if
    SINLON(i) = sin(ALON(i))
    COSLON(i) = cos(ALON(i))
  end do
 !$OMP END DO
!$OMP END PARALLEL

!  write(6,*) 'YLAT(:)=',YLAT(:)
!  write(6,*) 'WEIGHT(:)=',WEIGHT(:)

  call fft__ini( JCN_ZONALFILTER, M0, IMAX, JMAX, MMAX, coslat )

  XGAUSS(1) = -0.94910791234275852453d0
  XGAUSS(2) = -0.74153118559939443986d0
  XGAUSS(3) = -0.40584515137739716691d0
  XGAUSS(4) =  0.0d0
  XGAUSS(5) =  0.40584515137739716691d0
  XGAUSS(6) =  0.74153118559939443986d0
  XGAUSS(7) =  0.94910791234275852453d0

  WGAUSS(1) = 0.1294849661688696932d0
  WGAUSS(2) = 0.27970539148927666793d0
  WGAUSS(3) = 0.38183005050511894494d0
  WGAUSS(4) = 0.41795918367346938776d0
  WGAUSS(5) = 0.38183005050511894494d0
  WGAUSS(6) = 0.27970539148927666793d0
  WGAUSS(7) = 0.1294849661688696932d0

end subroutine initialize




!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&



subroutine main
!
  complex(8),save :: qff(NNUM,0:MMAX)
  complex(8),save :: qgg(NNUM,0:MMAX)
  complex(8),save :: qhh(NNUM,0:MMAX)
!
  real(8),save :: ff(IMAX,JMAX)
  real(8),save :: ff2(IMAX,JMAX)
  real(8),save :: lap_ff(IMAX,JMAX)
  real(8),save :: gg(IMAX,JMAX)
  real(8),save :: gg2(IMAX,JMAX)
  real(8),save :: hh(IMAX,JMAX)
  real(8),save :: hh2(IMAX,JMAX)
  real(8),save :: helm_hh(IMAX,JMAX)
  
  real(8),save :: ameanj(JMAX)
  real(8),save :: ameanj2(JMAX)
  
  real(8) :: alonc, alatc, rr, r, eps, ww1, ww2, ww3, ee, ee2
!
  integer :: iterx, kt_end, ntimes_hour, iadv_only
  integer :: kind_phi, kind_phis, kind_uv, nstep, nstep_start, intv_monit
  integer :: istep, i, j, nn, idiv, i1, i2, idepart_prevstep, ntmax
  integer :: istep_monit = 0
! =====================================================================
!
    alonc=3.0d0*PI/2                                                          
!    alonc=2.5*PI/2
                                                                                              
!    alatc=0.0d0
    alatc=PI/2-0.05d0
!    alatc=0.5*PI/2

    rr=ER/3
    eps=0.01d0*ER**2
    
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        r = ER*acos( sin(alatc)*SINLAT(j) + cos(alatc)*COSLAT(j)*cos(ALON(i)-alonc) )
        if ( r < rr ) then
          ww1 = PI*r/rr
          ww2 = PI*ER/rr
          ff(i,j) = 1000.0d0/4*( 1.0d0 + cos(ww1) )**2
          if ( r > 0.0d0 ) then  
            ww3 = -cos(r/ER)/sin(r/ER)*1000.0d0/(2*ER**2)*ww2*(1.0d0+cos(ww1))*sin(ww1)
          else
            ww3 = -1000.0d0*(PI/rr)**2
          end if
          gg(i,j) = ww3 + 1000.0d0/(2*ER**2)*ww2**2*( sin(ww1)**2 - (1.0d0+cos(ww1))*cos(ww1) )
        else
          ff(i,j) = 0.0d0
          gg(i,j) = 0.0d0
        end if
        hh(i,j) = ff(i,j) - eps*gg(i,j)
      end do
      
      do i=1,IMAX
        ff2(i,j) = ff(i,j)
        gg2(i,j) = gg(i,j)
        helm_hh(i,j) = hh(i,j)
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL

    
    call fft__towave &
     &( ff2     )    !INOUT
    call fft_y__g2w &
     &( ff2,        &!IN
     &  qff    )    !INOUT
    call fft_y__w2g &
     &( qff,       &!IN
     &  ff2    )    !OUT
    call fft__togrid &
     &( ff2    )    !INOUT

    call uv2divrot_dfs__laplacian(qff) !INOUT
    call fft_y__w2g &
     &( qff,       &!IN
     &  lap_ff    )  !OUT
    call fft__togrid &
     &( lap_ff    )  !INOUT
    
    call fft__towave &
     &( gg2     )    !INOUT
    call fft_y__g2w &
     &( gg2,        &!IN
     &  qgg    )    !INOUT
    call fft_y__w2g &
     &( qgg,       &!IN
     &  gg2    )    !OUT
    call fft__togrid &
     &( gg2    )    !INOUT

    call fft__towave &
     &( helm_hh  )    !INOUT
    call fft_y__g2w &
     &( helm_hh,    &!IN
     &  qhh    )     !INOUT
    call helmholtz_dfs__run &!harmonic spectral filter
     &( eps, eps,   &!IN
     &  qhh    )     !INOUT
    call fft_y__w2g &
     &( qhh,       &!IN
     &  helm_hh    )  !OUT
    call fft__togrid &
     &( helm_hh    )  !INOUT

!    call grads__outxy( "ff", 1, IMAX, JMAX, ylat, ff ) !IN
!    call grads__outxy( "ff2", 1, IMAX, JMAX, ylat, ff2 ) !IN
!    call grads__outxy( "gg", 1, IMAX, JMAX, ylat, gg ) !IN   
!    call grads__outxy( "gg2", 1, IMAX, JMAX, ylat, gg2 ) !IN     
!    call grads__outxy( "lap_ff", 1, IMAX, JMAX, ylat, lap_ff ) !IN
!    call grads__outxy( "hh", 1, IMAX, JMAX, ylat, hh ) !IN
!    call grads__outxy( "helm_hh", 1, IMAX, JMAX, ylat, helm_hh ) !IN
        
    ameanj(:) = 0.0d0
    ameanj2(:) = 0.0d0
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        ameanj(j) = ameanj(j) + ( lap_ff(i,j) - gg(i,j) )**2
        ameanj2(j) = ameanj2(j) + gg(i,j)**2
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL
    ee  = sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    ee2 = sqrt( sum(weight(:)*ameanj2(:))/( IMAX*sum(weight(:)) ) )
    write(6,*) "RMSE(lap_ff,gg_exact)=",ee,"(",ee/ee2,")"

    ameanj(:) = 0.0d0
    ameanj2(:) = 0.0d0
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        ameanj(j) = ameanj(j) + ( lap_ff(i,j) - gg2(i,j) )**2
        ameanj2(j) = ameanj2(j) + gg2(i,j)**2
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL
    ee  = sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    ee2 = sqrt( sum(weight(:)*ameanj2(:))/( IMAX*sum(weight(:)) ) )
    write(6,*) "RMSE(lap_ff,gg_trunc)=",ee,"(",ee/ee2,")"

!    ameanj(:) = 0.0d0
!   !$OMP PARALLEL default(SHARED), private(i,j,r)
!   !$OMP DO schedule(STATIC)
!    do j=1,JMAX
!      do i=1,IMAX
!        ameanj(j) = ameanj(j) + ( gg2(i,j) - gg(i,j) )**2
!      end do
!    end do
!   !$OMP END DO
!   !$OMP END PARALLEL
!    write(6,*) "RMSE(gg2,gg)=",sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    
    ameanj(:) = 0.0d0
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        ameanj(j) = ameanj(j) + lap_ff(i,j)
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL
    write(6,*) "mean(lap_ff)=",sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) )

    ameanj(:) = 0.0d0
    ameanj2(:) = 0.0d0
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        ameanj(j) = ameanj(j) + ( helm_hh(i,j) - ff(i,j) )**2
        ameanj2(j) = ameanj2(j) + ff(i,j)**2
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL
    ee  = sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    ee2 = sqrt( sum(weight(:)*ameanj2(:))/( IMAX*sum(weight(:)) ) )
    write(6,*) "RMSE(helm_hh,ff_true)=",ee,"(",ee/ee2,")"

    ameanj(:) = 0.0d0
    ameanj2(:) = 0.0d0
   !$OMP PARALLEL default(SHARED), private(i,j,r)
   !$OMP DO schedule(STATIC)
    do j=1,JMAX
      do i=1,IMAX
        ameanj(j) = ameanj(j) + ( helm_hh(i,j) - ff2(i,j) )**2
        ameanj2(j) = ameanj2(j) + ff2(i,j)**2
      end do
    end do
   !$OMP END DO
   !$OMP END PARALLEL
    ee  = sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    ee2 = sqrt( sum(weight(:)*ameanj2(:))/( IMAX*sum(weight(:)) ) )
    write(6,*) "RMSE(helm_hh,ff_trunc)=",ee,"(",ee/ee2,")"

!    ameanj(:) = 0.0d0
!   !$OMP PARALLEL default(SHARED), private(i,j,r)
!   !$OMP DO schedule(STATIC)
!    do j=1,JMAX
!      do i=1,IMAX
!        ameanj(j) = ameanj(j) + ( ff2(i,j) - ff(i,j) )**2
!      end do
!    end do
!   !$OMP END DO
!   !$OMP END PARALLEL
!    write(6,*) "RMSE(ff2,ff)=",sqrt( sum(weight(:)*ameanj(:))/( IMAX*sum(weight(:)) ) )
    
end subroutine main


!&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&


end program test_laplacian_helmholtz
